@@ -35,6 +35,7 @@ type opts struct {
 	pwd       string
 	app       string
 	overlay   []string
+	userns    bool
 }
 
 // imageExec can be used to run/exec/shell a Singularity image
@@ -75,6 +76,9 @@ func imageExec(t *testing.T, action string, opts opts, imagePath string, command
 	if opts.app != "" {
 		argv = append(argv, "--app", opts.app)
 	}
+	if opts.userns {
+		argv = append(argv, "-u")
+	}
 	argv = append(argv, imagePath)
 	argv = append(argv, command...)
 
@@ -176,8 +180,8 @@ func testSingularityExec(t *testing.T) {
 		{"ScifTestfolderOrg", appsImage, "exec", []string{"test", "-d", "/scif/apps/foo"}, opts{}, 0, true},
 		{"ScifTestfolderOrg", appsImage, "exec", []string{"test", "-d", "/scif/apps/bar"}, opts{}, 0, true},
 		// blocked by issue [scif-apps] Files created at install step fall into an unexpected path #2404
-		//{"ScifTestfolderOrg", appsImage, "exec", []string{"test", "-f", "/scif/apps/foo/filefoo.exec"}, opts{}, 0, true},
-		//{"ScifTestfolderOrg", appsImage, "exec", []string{"test", "-f", "/scif/apps/bar/filebar.exec"}, opts{}, 0, true},
+		{"ScifTestfolderOrg", appsImage, "exec", []string{"test", "-f", "/scif/apps/foo/filefoo.exec"}, opts{}, 0, true},
+		{"ScifTestfolderOrg", appsImage, "exec", []string{"test", "-f", "/scif/apps/bar/filebar.exec"}, opts{}, 0, true},
 		{"ScifTestfolderOrg", appsImage, "exec", []string{"test", "-d", "/scif/data/foo/output"}, opts{}, 0, true},
 		{"ScifTestfolderOrg", appsImage, "exec", []string{"test", "-d", "/scif/data/foo/input"}, opts{}, 0, true},
 		{"WorkdirContain", imagePath, "exec", []string{"test", "-f", tmpfile.Name()}, opts{workdir: "testdata", contain: true}, 0, false},
@@ -307,6 +311,13 @@ func testRunFromURI(t *testing.T) {
 		{"falseDocker", "docker://busybox:latest", "exec", []string{"false"}, opts{}, false},
 		{"falselibrary", "library://busybox:latest", "exec", []string{"false"}, opts{}, false},
 		{"falseShub", "shub://singularityhub/busybox", "exec", []string{"false"}, opts{}, false},
+		// exec from URI with user namespace enabled
+		{"trueDockerUserns", "docker://busybox:latest", "exec", []string{"true"}, opts{userns: true}, true},
+		{"trueLibraryUserns", "library://busybox:latest", "exec", []string{"true"}, opts{userns: true}, true},
+		{"trueShubUserns", "shub://singularityhub/busybox", "exec", []string{"true"}, opts{userns: true}, true},
+		{"falseDockerUserns", "docker://busybox:latest", "exec", []string{"false"}, opts{userns: true}, false},
+		{"falselibraryUserns", "library://busybox:latest", "exec", []string{"false"}, opts{userns: true}, false},
+		{"falseShubUserns", "shub://singularityhub/busybox", "exec", []string{"false"}, opts{userns: true}, false},
 	}
 
 	for _, tt := range tests {
