@@ -8,6 +8,9 @@ package cache
 import (
 	"os"
 	"path/filepath"
+
+	"github.com/sylabs/singularity/internal/pkg/sylog"
+	client "github.com/sylabs/singularity/pkg/client/library"
 )
 
 const (
@@ -29,12 +32,22 @@ func LibraryImage(sum, name string) string {
 
 // LibraryImageExists returns whether the image with the SHA sum exists in the LibraryImage cache
 func LibraryImageExists(sum, name string) (bool, error) {
-	_, err := os.Stat(LibraryImage(sum, name))
+	imagePath := LibraryImage(sum, name)
+	_, err := os.Stat(imagePath)
 	if os.IsNotExist(err) {
 		return false, nil
 	} else if err != nil {
 		return false, err
 	}
 
+	cacheSum, err := client.ImageHash(imagePath)
+	if err != nil {
+		return false, err
+	}
+	if cacheSum != sum {
+		sylog.Debugf("Cached File Sum(%s) and Expected Sum(%s) does not match", cacheSum, sum)
+		return false, nil
+	}
+
 	return true, nil
 }
