@@ -1,4 +1,4 @@
-// Copyright (c) 2018, Sylabs Inc. All rights reserved.
+// Copyright (c) 2018-2019, Sylabs Inc. All rights reserved.
 // This software is licensed under a 3-clause BSD license. Please consult the
 // LICENSE.md file distributed with the sources of this project regarding your
 // rights to use or distribute this software.
@@ -14,12 +14,11 @@ import (
 	"strings"
 	"syscall"
 
-	"github.com/opencontainers/runtime-spec/specs-go"
-	"github.com/sylabs/sif/pkg/sif"
+	"github.com/sylabs/singularity/internal/pkg/util/mainthread"
+
+	specs "github.com/opencontainers/runtime-spec/specs-go"
 	"github.com/sylabs/singularity/internal/pkg/buildcfg"
 	"github.com/sylabs/singularity/internal/pkg/cgroups"
-	"github.com/sylabs/singularity/internal/pkg/image"
-	"github.com/sylabs/singularity/internal/pkg/network"
 	"github.com/sylabs/singularity/internal/pkg/runtime/engines/singularity/rpc/client"
 	"github.com/sylabs/singularity/internal/pkg/sylog"
 	"github.com/sylabs/singularity/internal/pkg/util/fs"
@@ -28,12 +27,20 @@ import (
 	"github.com/sylabs/singularity/internal/pkg/util/fs/layout/layer/overlay"
 	"github.com/sylabs/singularity/internal/pkg/util/fs/layout/layer/underlay"
 	"github.com/sylabs/singularity/internal/pkg/util/fs/mount"
-	"github.com/sylabs/singularity/internal/pkg/util/fs/proc"
 	"github.com/sylabs/singularity/internal/pkg/util/user"
+	"github.com/sylabs/singularity/pkg/image"
+	"github.com/sylabs/singularity/pkg/network"
+	"github.com/sylabs/singularity/pkg/util/fs/proc"
 	"github.com/sylabs/singularity/pkg/util/loop"
 	"golang.org/x/crypto/ssh/terminal"
 )
 
+// defaultCNIConfPath is the default directory to CNI network configuration files
+var defaultCNIConfPath = filepath.Join(buildcfg.SYSCONFDIR, "singularity", "network")
+
+// defaultCNIPluginPath is the default directory to CNI plugins executables
+var defaultCNIPluginPath = filepath.Join(buildcfg.LIBEXECDIR, "singularity", "cni")
+
 type container struct {
 	engine           *EngineOperations
 	rpcOps           *client.RPC
@@ -48,6 +55,7 @@ type container struct {
 	ipcNS            bool
 	mountInfoPath    string
 	skippedMount     []string
+	checkDest        []string
 	suidFlag         uintptr
 	devSourcePath    string
 }
@@ -62,6 +70,7 @@ func create(engine *EngineOperations, rpcOps *client.RPC, pid int) error {
 		sessionFsType:    engine.EngineConfig.File.MemoryFSType,
 		mountInfoPath:    fmt.Sprintf("/proc/%d/mountinfo", pid),
 		skippedMount:     make([]string, 0),
+		checkDest:        make([]string, 0),
 		suidFlag:         syscall.MS_NOSUID,
 	}
 
@@ -153,10 +162,10 @@ func create(engine *EngineOperations, rpcOps *client.RPC, pid int) error {
 	}
 
 	sylog.Debugf("Chroot into %s\n", c.session.FinalPath())
-	_, err = c.rpcOps.Chroot(c.session.FinalPath(), true)
+	_, err = c.rpcOps.Chroot(c.session.FinalPath(), "pivot")
 	if err != nil {
 		sylog.Debugf("Fallback to move/chroot")
-		_, err = c.rpcOps.Chroot(c.session.FinalPath(), false)
+		_, err = c.rpcOps.Chroot(c.session.FinalPath(), "move")
 		if err != nil {
 			return fmt.Errorf("chroot failed: %s", err)
 		}
@@ -165,16 +174,24 @@ func create(engine *EngineOperations, rpcOps *client.RPC, pid int) error {
 	if c.netNS {
 		if os.Geteuid() == 0 && !c.userNS {
 			/* hold a reference to container network namespace for cleanup */
-			f, err := os.Open("/proc/" + strconv.Itoa(pid) + "/ns/net")
+			f, err := syscall.Open("/proc/"+strconv.Itoa(pid)+"/ns/net", os.O_RDONLY, 0)
 			if err != nil {
 				return fmt.Errorf("can't open network namespace: %s", err)
 			}
-			nspath := fmt.Sprintf("/proc/%d/fd/%d", os.Getpid(), f.Fd())
+			nspath := fmt.Sprintf("/proc/%d/fd/%d", os.Getpid(), f)
 			networks := strings.Split(engine.EngineConfig.GetNetwork(), ",")
 
-			cniPath := &network.CNIPath{
-				Conf:   engine.EngineConfig.File.CniConfPath,
-				Plugin: engine.EngineConfig.File.CniPluginPath,
+			cniPath := &network.CNIPath{}
+
+			if engine.EngineConfig.File.CniConfPath != "" {
+				cniPath.Conf = engine.EngineConfig.File.CniConfPath
+			} else {
+				cniPath.Conf = defaultCNIConfPath
+			}
+			if engine.EngineConfig.File.CniPluginPath != "" {
+				cniPath.Plugin = engine.EngineConfig.File.CniPluginPath
+			} else {
+				cniPath.Plugin = defaultCNIPluginPath
 			}
 
 			setup, err := network.NewSetup(networks, strconv.Itoa(pid), nspath, cniPath)
@@ -186,6 +203,8 @@ func create(engine *EngineOperations, rpcOps *client.RPC, pid int) error {
 				return fmt.Errorf("%s", err)
 			}
 
+			setup.SetEnvPath("/bin:/sbin:/usr/bin:/usr/sbin")
+
 			if err := setup.AddNetworks(); err != nil {
 				return fmt.Errorf("%s", err)
 			}
@@ -199,8 +218,8 @@ func create(engine *EngineOperations, rpcOps *client.RPC, pid int) error {
 	if os.Geteuid() == 0 {
 		path := engine.EngineConfig.GetCgroupsPath()
 		if path != "" {
-			name := strconv.Itoa(pid)
-			manager := &cgroups.Manager{Pid: pid, Name: name}
+			cgroupPath := filepath.Join("/singularity", strconv.Itoa(pid))
+			manager := &cgroups.Manager{Pid: pid, Path: cgroupPath}
 			if err := manager.ApplyFromFile(path); err != nil {
 				return fmt.Errorf("Failed to apply cgroups ressources restriction: %s", err)
 			}
@@ -217,68 +236,28 @@ func create(engine *EngineOperations, rpcOps *client.RPC, pid int) error {
 	return nil
 }
 
-func (c *container) setupWritableSIFImage(img *image.Image, overlayEnabled bool) error {
-	fimg, err := sif.LoadContainerFp(img.File, !img.Writable)
-	if err != nil {
-		return err
-	}
-
-	// Get the default system partition image
-	part, _, err := fimg.GetPartPrimSys()
-	if err != nil {
-		return err
-	}
-
-	// Get descriptor for group associated with system partition
-	descriptors, _, err := fimg.GetPartFromGroup(part.Groupid)
-	if err != nil {
-		return err
-	}
-
-	// Determine if an overlay partition exists
+func (c *container) setupSIFOverlay(img *image.Image, writable bool) error {
+	// Determine if overlay partitions exists
 	overlayPart := 0
-	for _, desc := range descriptors {
-		ptype, err := desc.GetPartType()
-		if err != nil {
-			return err
-		}
-		if ptype == sif.PartOverlay {
-			fstype, err := desc.GetFsType()
-			if err != nil {
-				continue
-			}
-			if fstype == sif.FsExt3 {
-				if overlayPart == 0 {
-					if !overlayEnabled {
-						return fmt.Errorf("Can't mount SIF image as read-write, overlay is not enabled")
-					}
+	overlayImg := c.engine.EngineConfig.GetOverlayImage()
+	imglist := c.engine.EngineConfig.GetImageList()
 
-					imgCopy := *img
-
-					imgCopy.Type = image.EXT3
-					imgCopy.Offset = uint64(desc.Fileoff)
-					imgCopy.Size = uint64(desc.Filelen)
-					imgCopy.RootFS = false
-					imgCopy.Writable = true
-
-					imglist := c.engine.EngineConfig.GetImageList()
-					imglist = append(imglist, imgCopy)
-
-					c.engine.EngineConfig.SetImageList(imglist)
-
-					overlayImg := c.engine.EngineConfig.GetOverlayImage()
-					overlayImg = append(overlayImg, imgCopy.Path)
-					c.engine.EngineConfig.SetOverlayImage(overlayImg)
-				}
-				overlayPart++
-			}
+	for _, p := range img.Partitions[1:] {
+		if p.Type == image.EXT3 || p.Type == image.SQUASHFS {
+			imgCopy := *img
+			imgCopy.Type = int(p.Type)
+			imgCopy.Partitions = []image.Section{p}
+			imglist = append(imglist, imgCopy)
+			overlayImg = append(overlayImg, imgCopy.Path)
+			overlayPart++
 		}
 	}
 
-	if overlayPart > 1 {
-		sylog.Warningf("more than one writable overlay partition found, taking the first")
-	} else if overlayPart == 0 {
-		return fmt.Errorf("no overlay partition found")
+	c.engine.EngineConfig.SetOverlayImage(overlayImg)
+	c.engine.EngineConfig.SetImageList(imglist)
+
+	if overlayPart == 0 && writable {
+		return fmt.Errorf("no SIF writable overlay partition found")
 	}
 
 	return nil
@@ -304,27 +283,32 @@ func (c *container) setupSessionLayout(system *mount.System) error {
 		}
 	}
 
-	if c.engine.EngineConfig.GetWritableImage() && !writableTmpfs {
-		imgObject, err := c.loadImage(c.engine.EngineConfig.GetImage(), true)
-		if err != nil {
-			return err
-		}
+	imgObject, err := c.loadImage(c.engine.EngineConfig.GetImage(), true)
+	if err != nil {
+		return fmt.Errorf("while loading image object: %s", err)
+	}
 
+	if c.engine.EngineConfig.GetWritableImage() && !writableTmpfs {
+		sylog.Debugf("Image is writable, not attempting to use overlay or underlay\n")
 		if imgObject.Type == image.SIF {
-			err = c.setupWritableSIFImage(imgObject, overlayEnabled)
+			err = c.setupSIFOverlay(imgObject, c.engine.EngineConfig.GetWritableImage())
 			if err == nil {
 				return c.setupOverlayLayout(system, sessionPath)
 			}
-			sylog.Warningf("%s", err)
-		} else {
-			sylog.Debugf("Image is writable, not attempting to use overlay or underlay\n")
+			sylog.Warningf("While attempting to set up SIFOverlay: %s", err)
 		}
-
 		return c.setupDefaultLayout(system, sessionPath)
 	}
 
 	if overlayEnabled {
 		sylog.Debugf("Attempting to use overlayfs (enable overlay = %v)\n", c.engine.EngineConfig.File.EnableOverlay)
+		if imgObject.Type == image.SIF {
+			err = c.setupSIFOverlay(imgObject, c.engine.EngineConfig.GetWritableImage())
+			if err == nil {
+				return c.setupOverlayLayout(system, sessionPath)
+			}
+			sylog.Warningf("While attempting to set up SIFOverlay: %s", err)
+		}
 		return c.setupOverlayLayout(system, sessionPath)
 	}
 
@@ -353,7 +337,7 @@ func (c *container) setupOverlayLayout(system *mount.System, sessionPath string)
 	}
 
 	c.sessionLayerType = "overlay"
-	return system.RunAfterTag(mount.LayerTag, c.setSlaveMount)
+	return system.RunAfterTag(mount.LayerTag, c.setPropagationMount)
 }
 
 // setupUnderlayLayout sets up the session with underlay "filesystem"
@@ -364,7 +348,7 @@ func (c *container) setupUnderlayLayout(system *mount.System, sessionPath string
 	}
 
 	c.sessionLayerType = "underlay"
-	return system.RunAfterTag(mount.LayerTag, c.setSlaveMount)
+	return system.RunAfterTag(mount.LayerTag, c.setPropagationMount)
 }
 
 // setupDefaultLayout sets up the session without overlay or underlay
@@ -375,7 +359,7 @@ func (c *container) setupDefaultLayout(system *mount.System, sessionPath string)
 	}
 
 	c.sessionLayerType = "none"
-	return system.RunAfterTag(mount.RootfsTag, c.setSlaveMount)
+	return system.RunAfterTag(mount.RootfsTag, c.setPropagationMount)
 }
 
 // isLayerEnabled returns whether or not overlay or underlay system
@@ -415,7 +399,7 @@ func (c *container) mount(point *mount.Point) error {
 	return nil
 }
 
-func (c *container) setSlaveMount(system *mount.System) error {
+func (c *container) setPropagationMount(system *mount.System) error {
 	pflags := uintptr(syscall.MS_REC)
 
 	if c.engine.EngineConfig.File.MountSlave {
@@ -472,14 +456,11 @@ func (c *container) mountGeneric(mnt *mount.Point) (err error) {
 	flags, opts := mount.ConvertOptions(mnt.Options)
 	optsString := strings.Join(opts, ",")
 	sessionPath := c.session.Path()
-	remount := false
+	remount := mount.HasRemountFlag(flags)
+	propagation := mount.HasPropagationFlag(flags)
 	source := mnt.Source
 	dest := ""
 
-	if flags&syscall.MS_REMOUNT != 0 {
-		remount = true
-	}
-
 	if flags&syscall.MS_BIND != 0 && !remount {
 		if _, err := os.Stat(source); os.IsNotExist(err) {
 			c.skippedMount = append(c.skippedMount, mnt.Destination)
@@ -505,7 +486,7 @@ func (c *container) mountGeneric(mnt *mount.Point) (err error) {
 		}
 	}
 
-	if remount {
+	if remount || propagation {
 		for _, skipped := range c.skippedMount {
 			if skipped == mnt.Destination {
 				return nil
@@ -513,19 +494,20 @@ func (c *container) mountGeneric(mnt *mount.Point) (err error) {
 		}
 		sylog.Debugf("Remounting %s\n", dest)
 	} else {
-		// detection of mounted points for underlay layer is not really a simple
-		// task, detection is disabled with this layer for the time being
-		if c.sessionLayerType != "underlay" {
-			mounted := c.checkMounted(dest)
-			if mounted != "" {
-				c.skippedMount = append(c.skippedMount, mnt.Destination)
-				sylog.Debugf("Skipping mount %s, %s already mounted", dest, mounted)
-				return nil
+		for _, d := range c.checkDest {
+			if d == mnt.Destination {
+				mounted := c.checkMounted(dest)
+				if mounted != "" {
+					c.skippedMount = append(c.skippedMount, mnt.Destination)
+					sylog.Debugf("Skipping mount %s, %s already mounted", dest, mounted)
+					return nil
+				}
+				break
 			}
 		}
 		sylog.Debugf("Mounting %s to %s\n", source, dest)
 
-		// in scontainer stage 1 we changed current working directory to
+		// in stage 1 we changed current working directory to
 		// sandbox image directory, just pass "." as source argument to
 		// be sure RPC mount the right sandbox image
 		if dest == c.session.RootFsPath() && flags&syscall.MS_BIND != 0 {
@@ -573,7 +555,8 @@ func (c *container) mountImage(mnt *mount.Point) error {
 		Flags:     loopFlags,
 	}
 
-	number, err := c.rpcOps.LoopDevice(mnt.Source, attachFlag, *info, maxDevices)
+	shared := c.engine.EngineConfig.File.SharedLoopDevices
+	number, err := c.rpcOps.LoopDevice(mnt.Source, attachFlag, *info, maxDevices, shared)
 	if err != nil {
 		return fmt.Errorf("failed to find loop device: %s", err)
 	}
@@ -591,31 +574,35 @@ func (c *container) mountImage(mnt *mount.Point) error {
 func (c *container) loadImage(path string, rootfs bool) (*image.Image, error) {
 	list := c.engine.EngineConfig.GetImageList()
 
+	if len(list) == 0 {
+		return nil, fmt.Errorf("no root filesystem found in %s", path)
+	}
+
 	if rootfs {
-		for _, img := range list {
-			if img.RootFS {
-				img.File = os.NewFile(img.Fd, img.Path)
-				if img.File == nil {
-					return nil, fmt.Errorf("can't find image %s", path)
-				}
+		img := list[0]
+		if img.File == nil {
+			return &img, nil
+		}
+		img.File = os.NewFile(img.Fd, img.Path)
+		if img.File == nil {
+			return nil, fmt.Errorf("can't find image %s", path)
+		}
+		return &img, nil
+	}
+	for _, img := range list[1:] {
+		p, err := image.ResolvePath(path)
+		if err != nil {
+			return nil, err
+		}
+		if p == img.Path {
+			if img.File == nil {
 				return &img, nil
 			}
-		}
-	} else {
-		for _, img := range list {
-			if !img.RootFS {
-				p, err := image.ResolvePath(path)
-				if err != nil {
-					return nil, err
-				}
-				if p == img.Path {
-					img.File = os.NewFile(img.Fd, img.Path)
-					if img.File == nil {
-						return nil, fmt.Errorf("can't find image %s", path)
-					}
-					return &img, nil
-				}
+			img.File = os.NewFile(img.Fd, img.Path)
+			if img.File == nil {
+				return nil, fmt.Errorf("can't find image %s", path)
 			}
+			return &img, nil
 		}
 	}
 
@@ -639,36 +626,10 @@ func (c *container) addRootfsMount(system *mount.System) error {
 	}
 
 	mountType := ""
+	offset := imageObject.Partitions[0].Offset
+	size := imageObject.Partitions[0].Size
 
-	switch imageObject.Type {
-	case image.SIF:
-		// Load the SIF file
-		fimg, err := sif.LoadContainerFp(imageObject.File, !imageObject.Writable)
-		if err != nil {
-			return err
-		}
-
-		// Get the default system partition image
-		part, _, err := fimg.GetPartPrimSys()
-		if err != nil {
-			return err
-		}
-
-		// record the fs type
-		fstype, err := part.GetFsType()
-		if err != nil {
-			return err
-		}
-		if fstype == sif.FsSquash {
-			mountType = "squashfs"
-		} else if fstype == sif.FsExt3 {
-			mountType = "ext3"
-		} else {
-			return fmt.Errorf("unknown file system type: %v", fstype)
-		}
-
-		imageObject.Offset = uint64(part.Fileoff)
-		imageObject.Size = uint64(part.Filelen)
+	switch imageObject.Partitions[0].Type {
 	case image.SQUASHFS:
 		mountType = "squashfs"
 	case image.EXT3:
@@ -686,7 +647,23 @@ func (c *container) addRootfsMount(system *mount.System) error {
 	}
 
 	sylog.Debugf("Mounting block [%v] image: %v\n", mountType, rootfs)
-	return system.Points.AddImage(mount.RootfsTag, imageObject.Source, c.session.RootFsPath(), mountType, flags, imageObject.Offset, imageObject.Size)
+	if err := system.Points.AddImage(
+		mount.RootfsTag,
+		imageObject.Source,
+		c.session.RootFsPath(),
+		mountType,
+		flags,
+		offset,
+		size,
+	); err != nil {
+		return err
+	}
+
+	if imageObject.Writable {
+		return system.Points.AddPropagation(mount.DevTag, c.session.RootFsPath(), syscall.MS_UNBINDABLE)
+	}
+
+	return nil
 }
 
 func (c *container) overlayUpperWork(system *mount.System) error {
@@ -775,6 +752,8 @@ func (c *container) addOverlayMount(system *mount.System) error {
 		nb++
 
 		src := imageObject.Source
+		offset := imageObject.Partitions[0].Offset
+		size := imageObject.Partitions[0].Size
 
 		switch imageObject.Type {
 		case image.EXT3:
@@ -785,14 +764,14 @@ func (c *container) addOverlayMount(system *mount.System) error {
 				ov.AddLowerDir(filepath.Join(dst, "upper"))
 			}
 
-			err = system.Points.AddImage(mount.PreLayerTag, src, dst, "ext3", flags, imageObject.Offset, imageObject.Size)
+			err = system.Points.AddImage(mount.PreLayerTag, src, dst, "ext3", flags, offset, size)
 			if err != nil {
-				return err
+				return fmt.Errorf("while adding ext3 image: %s", err)
 			}
 			flags &^= syscall.MS_RDONLY
 		case image.SQUASHFS:
 			flags := uintptr(c.suidFlag | syscall.MS_NODEV | syscall.MS_RDONLY)
-			err = system.Points.AddImage(mount.PreLayerTag, src, dst, "squashfs", flags, imageObject.Offset, imageObject.Size)
+			err = system.Points.AddImage(mount.PreLayerTag, src, dst, "squashfs", flags, offset, size)
 			if err != nil {
 				return err
 			}
@@ -805,7 +784,7 @@ func (c *container) addOverlayMount(system *mount.System) error {
 			flags := uintptr(c.suidFlag | syscall.MS_NODEV)
 			err = system.Points.AddBind(mount.PreLayerTag, imageObject.Path, dst, flags)
 			if err != nil {
-				return err
+				return fmt.Errorf("while adding sandbox image: %s", err)
 			}
 			system.Points.AddRemount(mount.PreLayerTag, dst, flags)
 
@@ -820,6 +799,11 @@ func (c *container) addOverlayMount(system *mount.System) error {
 			return fmt.Errorf("unknown image format")
 		}
 
+		err = system.Points.AddPropagation(mount.DevTag, dst, syscall.MS_UNBINDABLE)
+		if err != nil {
+			return err
+		}
+
 		if imageObject.Writable && !hasUpper {
 			upper := filepath.Join(dst, "upper")
 			work := filepath.Join(dst, "work")
@@ -841,7 +825,7 @@ func (c *container) addOverlayMount(system *mount.System) error {
 		}
 	}
 
-	return nil
+	return system.Points.AddPropagation(mount.DevTag, c.session.FinalPath(), syscall.MS_UNBINDABLE)
 }
 
 func (c *container) addKernelMount(system *mount.System) error {
@@ -864,6 +848,7 @@ func (c *container) addKernelMount(system *mount.System) error {
 		if err != nil {
 			return fmt.Errorf("unable to add proc to mount list: %s", err)
 		}
+		sylog.Verbosef("Default mount: /proc:/proc")
 	} else {
 		sylog.Verbosef("Skipping /proc mount")
 	}
@@ -884,6 +869,7 @@ func (c *container) addKernelMount(system *mount.System) error {
 		if err != nil {
 			return fmt.Errorf("unable to add sys to mount list: %s", err)
 		}
+		sylog.Verbosef("Default mount: /sys:/sys")
 	} else {
 		sylog.Verbosef("Skipping /sys mount")
 	}
@@ -1024,7 +1010,7 @@ func (c *container) addDevMount(system *mount.System) error {
 			//  and also check the device that docker uses,
 			//  /dev/console.
 			procfd := fmt.Sprintf("/proc/self/fd/%d", fd)
-			ttylink, err := os.Readlink(procfd)
+			ttylink, err := mainthread.Readlink(procfd)
 			if err != nil {
 				return err
 			}
@@ -1108,6 +1094,7 @@ func (c *container) addDevMount(system *mount.System) error {
 		if err != nil {
 			return fmt.Errorf("unable to add dev to mount list: %s", err)
 		}
+		sylog.Verbosef("Default mount: /dev:/dev")
 	} else if c.engine.EngineConfig.File.MountDev == "no" {
 		sylog.Verbosef("Not mounting /dev inside the container, disallowed by configuration")
 	}
@@ -1347,11 +1334,14 @@ func (c *container) addUserbindsMount(system *mount.System) error {
 
 		sylog.Debugf("Adding %s to mount list\n", src)
 
-		if err := system.Points.AddBind(mount.UserbindsTag, src, dst, flags); err != nil {
-			return fmt.Errorf("unabled to %s to mount list: %s", src, err)
+		if err := system.Points.AddBind(mount.UserbindsTag, src, dst, flags); err != nil && err == mount.ErrMountExists {
+			sylog.Warningf("destination %s already in mount list: %s", src, err)
+		} else if err != nil {
+			return fmt.Errorf("unable to add %s to mount list: %s", src, err)
+		} else {
+			system.Points.AddRemount(mount.UserbindsTag, dst, flags)
+			flags &^= syscall.MS_RDONLY
 		}
-		system.Points.AddRemount(mount.UserbindsTag, dst, flags)
-		flags &^= syscall.MS_RDONLY
 	}
 
 	sylog.Debugf("Checking for 'user bind control' in configuration file")
@@ -1379,20 +1369,20 @@ func (c *container) addTmpMount(system *mount.System) error {
 				return nil
 			}
 
-			vartmpSource = "/var_tmp"
+			vartmpSource = "var_tmp"
 
 			workdir, err := filepath.Abs(filepath.Clean(workdir))
 			if err != nil {
 				sylog.Warningf("Can't determine absolute path of workdir %s", workdir)
 			}
 
-			tmpSource = workdir + tmpSource
-			vartmpSource = workdir + vartmpSource
+			tmpSource = filepath.Join(workdir, tmpSource)
+			vartmpSource = filepath.Join(workdir, vartmpSource)
 
-			if err := fs.MkdirAll(tmpSource, 0755); err != nil {
+			if err := fs.Mkdir(tmpSource, os.ModeSticky|0777); err != nil && !os.IsExist(err) {
 				return fmt.Errorf("failed to create %s: %s", tmpSource, err)
 			}
-			if err := fs.MkdirAll(vartmpSource, 0755); err != nil {
+			if err := fs.Mkdir(vartmpSource, os.ModeSticky|0777); err != nil && !os.IsExist(err) {
 				return fmt.Errorf("failed to create %s: %s", vartmpSource, err)
 			}
 		} else {
@@ -1400,11 +1390,17 @@ func (c *container) addTmpMount(system *mount.System) error {
 				if err := c.session.AddDir(tmpSource); err != nil {
 					return err
 				}
+				if err := c.session.Chmod(tmpSource, os.ModeSticky|0777); err != nil {
+					return err
+				}
 			}
 			if _, err := c.session.GetPath(vartmpSource); err != nil {
 				if err := c.session.AddDir(vartmpSource); err != nil {
 					return err
 				}
+				if err := c.session.Chmod(vartmpSource, os.ModeSticky|0777); err != nil {
+					return err
+				}
 			}
 			tmpSource, _ = c.session.GetPath(tmpSource)
 			vartmpSource, _ = c.session.GetPath(vartmpSource)
@@ -1414,11 +1410,13 @@ func (c *container) addTmpMount(system *mount.System) error {
 
 	if err := system.Points.AddBind(mount.TmpTag, tmpSource, "/tmp", flags); err == nil {
 		system.Points.AddRemount(mount.TmpTag, "/tmp", flags)
+		sylog.Verbosef("Default mount: /tmp:/tmp")
 	} else {
 		return fmt.Errorf("could not mount container's /tmp directory: %s %s", err, tmpSource)
 	}
 	if err := system.Points.AddBind(mount.TmpTag, vartmpSource, "/var/tmp", flags); err == nil {
 		system.Points.AddRemount(mount.TmpTag, "/var/tmp", flags)
+		sylog.Verbosef("Default mount: /var/tmp:/var/tmp")
 	} else {
 		return fmt.Errorf("could not mount container's /var/tmp directory: %s", err)
 	}
@@ -1457,11 +1455,11 @@ func (c *container) addScratchMount(system *mount.System) error {
 
 		if hasWorkdir {
 			fullSourceDir = filepath.Join(sourceDir, filepath.Base(dir))
-			if err := fs.MkdirAll(fullSourceDir, 0750); err != nil {
+			if err := fs.MkdirAll(fullSourceDir, 0750); err != nil && !os.IsExist(err) {
 				return fmt.Errorf("could not create scratch working directory %s: %s", sourceDir, err)
 			}
 		} else {
-			src := filepath.Join("/scratch", filepath.Base(dir))
+			src := filepath.Join("/scratch", dir)
 			if err := c.session.AddDir(src); err != nil {
 				return fmt.Errorf("could not create scratch working directory %s: %s", sourceDir, err)
 			}
@@ -1492,7 +1490,11 @@ func (c *container) addCwdMount(system *mount.System) error {
 	}
 	cwd = c.engine.EngineConfig.OciConfig.Process.Cwd
 	if err := os.Chdir(cwd); err != nil {
-		sylog.Warningf("Could not set container working directory %s: %s", cwd, err)
+		if os.IsNotExist(err) {
+			sylog.Debugf("Container working directory %s doesn't exist, will retry after chroot", cwd)
+		} else {
+			sylog.Warningf("Could not set container working directory %s: %s", cwd, err)
+		}
 		return nil
 	}
 	current, err := os.Getwd()
@@ -1500,7 +1502,7 @@ func (c *container) addCwdMount(system *mount.System) error {
 		return fmt.Errorf("could not obtain current directory path: %s", err)
 	}
 	switch current {
-	case "/", "/etc", "/bin", "/mnt", "/usr", "/var", "/opt", "/sbin":
+	case "/", "/etc", "/bin", "/mnt", "/usr", "/var", "/opt", "/sbin", "/lib", "/lib64":
 		sylog.Verbosef("Not mounting CWD within operating system directory: %s", current)
 		return nil
 	}
@@ -1511,6 +1513,8 @@ func (c *container) addCwdMount(system *mount.System) error {
 	flags := uintptr(syscall.MS_BIND | c.suidFlag | syscall.MS_NODEV | syscall.MS_REC)
 	if err := system.Points.AddBind(mount.CwdTag, current, cwd, flags); err == nil {
 		system.Points.AddRemount(mount.CwdTag, cwd, flags)
+		c.checkDest = append(c.checkDest, cwd)
+		sylog.Verbosef("Default mount: %v: to the container", cwd)
 	} else {
 		sylog.Warningf("Could not bind CWD to container %s: %s", current, err)
 	}
@@ -1603,6 +1607,7 @@ func (c *container) addIdentityMount(system *mount.System) error {
 				if err != nil {
 					return fmt.Errorf("unable to add /etc/passwd to mount list: %s", err)
 				}
+				sylog.Verbosef("Default mount: /etc/passwd:/etc/passwd")
 			}
 		}
 	} else {
@@ -1625,6 +1630,7 @@ func (c *container) addIdentityMount(system *mount.System) error {
 			if err != nil {
 				return fmt.Errorf("unable to add /etc/group to mount list: %s", err)
 			}
+			sylog.Verbosef("Default mount: /etc/group:/etc/group")
 		}
 	} else {
 		sylog.Verbosef("Skipping bind of the host's /etc/group")
@@ -1668,6 +1674,7 @@ func (c *container) addResolvConfMount(system *mount.System) error {
 		if err != nil {
 			return fmt.Errorf("unable to add %s to mount list: %s", resolvConf, err)
 		}
+		sylog.Verbosef("Default mount: /etc/resolv.conf:/etc/resolv.conf")
 	} else {
 		sylog.Verbosef("Skipping bind of the host's %s", resolvConf)
 	}
@@ -1696,6 +1703,7 @@ func (c *container) addHostnameMount(system *mount.System) error {
 			if err != nil {
 				return fmt.Errorf("unable to add %s to mount list: %s", hostnameFile, err)
 			}
+			sylog.Verbosef("Default mount: /etc/hostname:/etc/hostname")
 			if _, err := c.rpcOps.SetHostname(hostname); err != nil {
 				return fmt.Errorf("failed to set container hostname: %s", err)
 			}
