@@ -1,4 +1,4 @@
-// Copyright (c) 2018, Sylabs Inc. All rights reserved.
+// Copyright (c) 2018-2019, Sylabs Inc. All rights reserved.
 // This software is licensed under a 3-clause BSD license. Please consult the
 // LICENSE.md file distributed with the sources of this project regarding your
 // rights to use or distribute this software.
@@ -7,12 +7,13 @@ package singularity
 
 import (
 	"github.com/sylabs/singularity/internal/pkg/runtime/engines/config"
+	singularityConfig "github.com/sylabs/singularity/internal/pkg/runtime/engines/singularity/config"
 )
 
 // EngineOperations describes a runtime engine
 type EngineOperations struct {
-	CommonConfig *config.Common `json:"-"`
-	EngineConfig *EngineConfig  `json:"engineConfig"`
+	CommonConfig *config.Common                  `json:"-"`
+	EngineConfig *singularityConfig.EngineConfig `json:"engineConfig"`
 }
 
 // InitConfig stores the pointer to config.Common
