@@ -6,6 +6,7 @@
 package mount
 
 import (
+	"fmt"
 	"syscall"
 	"testing"
 
@@ -59,7 +60,9 @@ func TestImage(t *testing.T) {
 		t.Errorf("should have failed with 0 size limit")
 	}
 	if err := points.AddImage(RootfsTag, "/fake", "/squash", "squashfs", 0, 0, 10); err == nil {
-		t.Errorf("should have failed with destination already in list")
+		t.Errorf("nil error returned, should have returned non-nil mount.ErrMountExists")
+	} else if err != nil && err != ErrMountExists {
+		t.Errorf("non-nil error should have been mount.ErrMountExists")
 	}
 	points.RemoveAll()
 
@@ -290,6 +293,25 @@ func TestRemount(t *testing.T) {
 	points.RemoveAll()
 }
 
+func TestAddPropagation(t *testing.T) {
+	test.DropPrivilege(t)
+	defer test.ResetPrivilege(t)
+
+	points := &Points{}
+
+	if err := points.AddPropagation(UserbindsTag, "", 0); err == nil {
+		t.Errorf("should have failed with empty destination")
+	}
+	if err := points.AddPropagation(UserbindsTag, "/mnt", 0); err == nil {
+		t.Errorf("should have failed with no propagation flag found")
+	}
+	if err := points.AddPropagation(UserbindsTag, "/mnt", syscall.MS_SHARED|syscall.MS_REC); err != nil {
+		t.Error(err)
+	}
+
+	points.RemoveAll()
+}
+
 func TestImport(t *testing.T) {
 	test.DropPrivilege(t)
 	defer test.ResetPrivilege(t)
@@ -314,15 +336,7 @@ func TestImport(t *testing.T) {
 					Source:      "/",
 					Destination: "/mnt",
 					Type:        "",
-					Options:     []string{"rbind"},
-				},
-			},
-			{
-				Mount: specs.Mount{
-					Source:      "",
-					Destination: "/mnt",
-					Type:        "",
-					Options:     []string{"rbind", "nosuid", "remount"},
+					Options:     []string{"rbind", "nosuid"},
 				},
 			},
 		},
@@ -391,7 +405,7 @@ func TestImport(t *testing.T) {
 		t.Errorf("wrong number of overlay mount point found")
 	}
 	bind := points.GetAllBinds()
-	if len(bind) != 2 {
+	if len(bind) != 1 {
 		t.Errorf("wrong number of bind mount point found")
 	}
 	fs := points.GetAllFS()
@@ -449,14 +463,6 @@ func TestImport(t *testing.T) {
 
 	invalidImport := map[AuthorizedTag][]Point{
 		UserbindsTag: {
-			{
-				Mount: specs.Mount{
-					Source:      "/",
-					Destination: "/mnt",
-					Type:        "",
-					Options:     []string{"rbind"},
-				},
-			},
 			{
 				Mount: specs.Mount{
 					Source:      "",
@@ -492,8 +498,9 @@ func TestImport(t *testing.T) {
 		t.Fatalf("returned a wrong number of mount points %d instead of 1", len(tmp))
 	}
 	hasContext := false
+	context := fmt.Sprintf("context=%q", mountLabel)
 	for _, option := range tmp[0].Options {
-		if option == "context="+mountLabel {
+		if option == context {
 			hasContext = true
 		}
 	}
@@ -540,13 +547,7 @@ func TestImport(t *testing.T) {
 			Source:      "/",
 			Destination: "/mnt",
 			Type:        "",
-			Options:     []string{"rbind"},
-		},
-		{
-			Source:      "",
-			Destination: "/mnt",
-			Type:        "",
-			Options:     []string{"rbind", "nosuid", "remount"},
+			Options:     []string{"rbind", "nosuid", "rshared"},
 		},
 		{
 			Source:      "",
@@ -576,8 +577,11 @@ func TestImport(t *testing.T) {
 	if err := points.ImportFromSpec(validSpecs); err != nil {
 		t.Error(err)
 	}
-	if len(points.GetAll()) != 5 {
-		t.Errorf("returned a wrong number of mount points %d instead of 5", len(points.GetAll()))
+	if len(points.GetByTag(KernelTag)) != 4 {
+		t.Errorf("returned a wrong number of mount kernel mount points %d instead of 4", len(points.GetByTag(KernelTag)))
+	}
+	if len(points.GetByTag(UserbindsTag)) != 3 {
+		t.Errorf("returned a wrong number of mount kernel mount points %d instead of 3", len(points.GetByTag(UserbindsTag)))
 	}
 	points.RemoveAll()
 
@@ -616,55 +620,3 @@ func TestTag(t *testing.T) {
 		}
 	}
 }
-
-func TestPaths(t *testing.T) {
-	test.DropPrivilege(t)
-	defer test.ResetPrivilege(t)
-
-	points := &Points{}
-	maskedPaths := []string{
-		"/proc/kcore",
-		"/proc/latency_stats",
-	}
-	roPaths := []string{
-		"/proc/asound",
-		"/proc/bus",
-	}
-	if err := points.AddMaskedPaths(maskedPaths); err != nil {
-		t.Error(err)
-	}
-	l := len(points.GetAll())
-	if l != 1 {
-		t.Errorf("returned a wrong number of mount points %d instead of 1", l)
-	}
-	l = len(points.GetByTag(OtherTag))
-	if l != 2 {
-		t.Errorf("returned a wrong number of mount points %d instead of 2", l)
-	}
-	points.RemoveAll()
-	if err := points.AddReadonlyPaths(roPaths); err != nil {
-		t.Error(err)
-	}
-	l = len(points.GetAll())
-	if l != 1 {
-		t.Errorf("returned a wrong number of mount points %d instead of 1", l)
-	}
-	l = len(points.GetByTag(OtherTag))
-	if l != 4 {
-		t.Errorf("returned a wrong number of mount points %d instead of 4", l)
-	}
-	maskedPaths = []string{
-		"/proc/kcore",
-		"/proc/kcore",
-	}
-	if err := points.AddMaskedPaths(maskedPaths); err == nil {
-		t.Errorf("should have failed with duplicated mount point")
-	}
-	roPaths = []string{
-		"/proc/asound",
-		"/proc/asound",
-	}
-	if err := points.AddReadonlyPaths(roPaths); err == nil {
-		t.Errorf("should have failed with duplicated mount point")
-	}
-}
